"Python3 (categorical) keyword finder."

##  NEWQUAY :
##
##  Categorical keyness calculation program :
##  Copyright 2013, R.S. Forsyth.
##  first version : 15/08/2013
##  last revision : 26/03/2015


import  math, os, string, sys
import  random, time
##  import  copy


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  add2dic
from quaylib import  fivenum, midvalue
##from statlib import  g2
##from statlib import  hedgesg
##from statlib import  superior
from quaylib import  addsnips
from quaylib import  getkeys, keyshow
from quaylib import  getfrex
from quaylib import  outpars


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","newquay"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfnum initial
    ("randseed",1789,1,999999),  ##  random seed, for replicability
    ("mindocf",3,0,10),  ##  minimum worthwhile document frequency
    ("topkeys",32,1,256),  ##  number of high-scoring keys to keep
    ("snipsize",115,4,2048),  ##  number of tokens in a snippet (57=ssss)
    ("keymode","g2"),  ##  choice of keyness function :
    ##  coda, doczeta, g2, hg, neozeta, roar, tuki.
    ("metafile",''),  ##  training corpus
    ("trainmet",''),  ##  ditto (alternative)
    ("dumpfile",''),  ##  machine-readable results file
    ("listfile",''),  ##  human-readable results file
    ("keyfile",''),  ##  file for o/p of keys
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"newquay.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "3.1"
Gold = (1.0+5.0**0.5)*0.5
Mindocs = 4
Dumptail = "_dump"
Listtail = "_pars"
Logtail = "_klog"
Keytail = "_keys"
Longname = 28  ##  longest class name; longer will be truncated
Maxtoks = 16 * 1024  ##  will only seek tokens in front chunk of text file


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if not paradat.trainmet:
        paradat.trainmet = paradat.metafile
    elif not paradat.metafile:
        paradat.metafile = paradat.trainmet
    assert  paradat.metafile == paradat.trainmet
    if paradat.keymode.startswith("hg"):
        paradat.keymode = "hedg"  ##  standard form of Hedges's g
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of main (training) corpus :
    metaname = getattr(paradat,"metafile",'')
    if not metaname:
        print("Can't proceed without metafile!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes)
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
    else:
        print("No. of text types =",paradat.cats)

    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    paradat.midsize = midvalue(paradat.sizelist)
    paradat.minsize = min(paradat.sizelist)
    print("Shortest =",paradat.minsize,"tokens.")
    print("median size =",paradat.midsize)
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("mean size =",round(paradat.meansize,2))
    print("Longest =",paradat.maxsize,"tokens.")
    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -5
    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].tottoks)
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  reference class is largest
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)
    print("reference category : ",paradat.refcode)

    for c in paradat.catlist:
        print(c,catinfo[c].docs)

    ##  attach extra attributes to doclist :
    paradat.totsnips = addsnips(paradat,doclist)
    for d in doclist:
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  attach category number as well as class name.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)  ##  each text has freqtab attached
        d.vocsize = len(d.freqtab)
        ##  also make overall freqtabs :
        ##  add2dic(allfrex,d.freqtab)
        ##  never used ??
    print("snippets =",paradat.totsnips)
    for j in []:  ##  sanity check
        d = doclist[j]
        print(d.dnum,d.name,d.outcome,d.size,d.vocsize,d.snippets)
        for j in [0,-1]:
            s = d.snipsets[j]
            print(len(s),sorted(s))

    random.seed(paradat.randseed)  ##  ensure repeatability (cd be a user parameter ?)

    ##  prepare o/p files :
    if paradat.outfile:
        logfile = paradat.outfile
    else:
        logfile = paradat.outpath + paradat.jobname + Logtail + ".txt"
    paradat.fo = open(logfile,'w',encoding="utf8")
    paradat.logfile = paradat.fo  ##  synomym
    parashow(paradat,paradat.fo)
    if paradat.dumpfile:
        dumpname = paradat.dumpfile
    else:
        dumpname = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    paradat.dumpfile = open(dumpname,'w',encoding="utf8")
    ##  also need separate results file :
    if paradat.listfile:
        listfile = paradat.listfile
    else:
        listfile = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")
    ##  also need file for o/p of keys :
    if paradat.keyfile:
        keyfile = paradat.keyfile
    else:
        keyfile = paradat.outpath + paradat.jobname + Keytail + ".txt"
    paradat.keyfile = open(keyfile,'w',encoding="utf8")

    ##  initial checking :
    ##  outpars(paradat)
    ##  print(type(paradat))  ##  needs str() to make it a string
    ##  print(isinstance(paradat,type(Sack())))  ##  => True
    indices = list(range(paradat.docs))  ##  all indices

    ##  find keys :
    keydata = []
    print("extracting keys :")
    for catname in paradat.catlist:
        print(catname)
        keydata.append(getkeys(paradat,doclist,catname,indices))
    outpars(paradat,paradat.listfile)
    
    ##  show keys :
    print(paradat.dateline,file=paradat.keyfile)
    print("parafile:",paradat.id,file=paradat.keyfile)
    print("metafile:",paradat.metafile,file=paradat.keyfile)
    print("keymode :",paradat.keymode,file=paradat.keyfile)
    print(paradat.docs,paradat.cats,file=paradat.keyfile)
    headline = "``                   rank    quayval   #self #other    %self   %other %midrate"
    for j,k in enumerate(keydata):
        ##  if paradat.cats <= 2 and k.catname != paradat.refcode:
            ##  continue  ##  with dichotomies, only show larger class
        print(j,k.catname)
        keyshow(k,paradat.dumpfile)
        nontoks = paradat.tottoks - catinfo[k.catname].tottoks
        lout = "{0:2d} {1:4d}        {2:8d}  {3:9d}".format(j,catinfo[k.catname].docs,
                                              catinfo[k.catname].tottoks,nontoks)
        print(NL+str(k.catname),file=paradat.keyfile)
        print("##"+lout,file=paradat.keyfile)
        print(headline,file=paradat.keyfile)
        listkeys(k,paradat.keyfile)
    
    ##  finale :
    paradat.dumpfile.close()
    paradat.listfile.close()
    paradat.logfile.close()
    paradat.keyfile.close()
    print("keys written onto file",os.path.abspath(paradat.keyfile.name))
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    
    return  0



##  LISTKEYS :
def listkeys (keydata,keyfile):
    "shows keys for a single class."

    ##  heading done already.
    nk = len(keydata.negkeys)
    for j,p in enumerate(keydata.poskeys):
        lout = "{0:20s} {1:4d} {2:10.4f}  {3:6d} {4:6d} {5:8.4f} {6:8.4f} {7:8.4f}". \
               format(p.item,j+1,p.quayval,p.rawfrex[0],p.rawfrex[2],
                100.0*p.r1,100.0*p.r2,100.0*p.midrate)
        print(lout,file=keyfile)
    for j,n in enumerate(keydata.negkeys):
        lout = "{0:20s} {1:4d} {2:10.4f}  {3:6d} {4:6d} {5:8.4f} {6:8.4f} {7:8.4f}". \
               format(n.item,j-nk,n.quayval,n.rawfrex[0],n.rawfrex[2],
                100.0*n.r1,100.0*n.r2,100.0*n.midrate)
        print(lout,file=keyfile)


    return




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
